"""
chrono24/session
~~~~~~~~~~~~~~~~
"""

import requests
import tenacity
from bs4 import BeautifulSoup

from chrono24.exceptions import RequestException

HEADERS = {
    "accept": "*/*",
    "accept-encoding": "gzip, deflate",
    "accept-language": "pt-PT,pt;q=0.9,en-US;q=0.8,en;q=0.7",
    "cookie": "uid=9c718efe-dcca-4e71-b92d-c3dd7b7f06cc",
    "referer": "https://a3853408329f84107a5d2b90c11d7c4b.safeframe.googlesyndication.com/",
    "sec-ch-ua": '" Not A;Brand";v="99", "Chromium";v="98", "Google Chrome";v="98"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '"Windows"',
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "user-agent": "Mozilla/5.0 (Windows NT 8.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.82 Safari/537.36",
}


def _log_retry_attempt(retry_state):
    """Logs retry state to console if request retry failed.

    Args:
        retry_state (tenacity.retry): Retry object.
    """
    print(f"Retrying request... Attempt # {retry_state.attempt_number}.")


RETRY_ARGS = {
    "wait": tenacity.wait.wait_random_exponential(multiplier=1, exp_base=2),
    "retry": tenacity.retry_if_exception_type(
        (
            requests.ConnectionError,
            requests.HTTPError,
            requests.RequestException,
        )
    ),
    "before_sleep": _log_retry_attempt,
}


def get_html(url, error_message="Failed to execute request.", max_attempts=8):
    """Takes URL and converts to `bs4.BeautifulSoup` object. Retries requests for `max_attempts` if request fails;
    if request still fails thereafter, None is returned.

    Args:
        url (str): URL to query and convert to `bs4.BeautifulSoup` object.
        error_message (str, optional): Error message to print if requests failed. Defaults to "Failed to
            execute request."
        max_attempts (int, optional): Maximum number of re-attempts prior to forfeiting request. Defaults
            to 8.

    Returns:
        bs4.BeautifulSoup: `bs4.BeautifulSoup` object if requested response was successful.

    Raises:
        RequestException: Repeated requests failed altogether.
    """
    response = get_response(url, error_message=error_message, max_attempts=max_attempts)
    return BeautifulSoup(response.text, "html.parser")


def get_response(*args, error_message="Failed to execute request.", max_attempts=8, **kwargs):
    """Makes get request using the tenacity wrapper to retry failed requests. Returns response object if
    request is successful; returns None if request failed after max retries.

    Args:
        *args (Any): Positional arguments to pass to `requests.Session().get`.
        error_message (str, optional): Error message to print if requests failed. Defaults to "Failed to
            execute request."
        max_attempts (int, optional): Maximum number of re-attempts prior to forfeiting request. Defaults
            to 8.
        **kwargs (Any): Keyword arguments to pass to `requests.Session().get`.

    Returns:
        requests.models.Response: Returns response if request was successful.

    Raises:
        RequestException: Repeated requests failed altogether.
    """
    try:
        return _get_tenacity_wrapped_response(*args, max_attempts=max_attempts, **kwargs)
    except (tenacity.RetryError, requests.exceptions.HTTPError) as e:
        print(error_message)
        raise RequestException from e


def _get_tenacity_wrapped_response(*args, max_attempts=8, **kwargs):
    """Makes tenacity-wrapped get request and retries request if it fails.

    Args:
        *args (Any): Positional arguments to pass to `requests.Session().get`.
        max_attempts (int): Maximum number of re-attempts prior to forfeiting request. Defaults to 8.
        **kwargs (Any): Keyword arguments to pass to `requests.Session().get`.

    Returns:
        requests.models.Response : Returns response and session object if request was successful.

    Raises:
        tenacity.RetryError: Failed to get request after max retries.
        requests.exceptions.HTTPError: An HTTP error occurred.
    """
    retry_args = RETRY_ARGS.copy()
    retry_args["stop"] = tenacity.stop.stop_after_attempt(max_attempts)
    for attempt in tenacity.Retrying(**retry_args):
        with attempt:
            response = requests.get(
                *args, **kwargs, headers={**HEADERS, **kwargs.get("headers", {})}
            )
            response.raise_for_status()
            return response



"""
tests/test_api
~~~~~~~~~~~~~~
"""

from functools import reduce

from pytest import mark

import chrono24


# fmt: off
# Fetch listings up-front to avoid flooding Chrono24 with requests
ROLEX_DJ = chrono24(query="Rolex DateJust")
SEARCH_LIMIT = 3
LISTINGS = list(ROLEX_DJ.search(limit=SEARCH_LIMIT))
DETAILED_LISTINGS = list(ROLEX_DJ.search_detail(limit=SEARCH_LIMIT))
JOINED_LISTINGS = (LISTINGS, DETAILED_LISTINGS)
LISTING = LISTINGS[0]
DETAILED_LISTING = DETAILED_LISTINGS[0]
JOINED_LISTING = (LISTING, DETAILED_LISTING)
STANDARD_LISTING_KEYS = {
    'id', 'url', 'manufacturer', 'certification_status',
    'title', 'description', 'price', 'shipping_price',
    'location', 'merchant_name', 'badge', 'image_urls',
}
# fmt: on


def test_attrs():
    """Test for checking attributes."""
    assert isinstance(ROLEX_DJ.count, int)
    assert isinstance(ROLEX_DJ.url, str)


@mark.parametrize("listings", JOINED_LISTINGS)
def test_search_limit_reached(listings):
    """Test for checking search limit."""
    assert len(listings) == SEARCH_LIMIT


def test_standard_listing_keys():
    """Ensure that the retrieved listing dictionary contains only the expected keys."""
    assert len(STANDARD_LISTING_KEYS.union(LISTING.keys())) == len(STANDARD_LISTING_KEYS)
    assert len(STANDARD_LISTING_KEYS.intersection(LISTING.keys())) == len(STANDARD_LISTING_KEYS)


def test_detailed_listing_keys():
    """Ensure that detailed search results have more keys than standard search results."""
    assert len(DETAILED_LISTING.keys()) > len(STANDARD_LISTING_KEYS)


@mark.parametrize("listing", JOINED_LISTING)
def test_listing_values_are_str(listing):
    """Ensure that values in a listing are of types `str` or `list` and contain strings."""
    listing_value_types = set(map(lambda x: type(x), listing.values()))
    num_listing_value_types = len(listing_value_types)
    # Assert only two types in listing values
    assert num_listing_value_types == 2
    # Assert the two value types are list and str
    assert len(set(listing_value_types.union((list, str)))) == num_listing_value_types
    # Assert values inside listing values are of type str
    list_values = [v for v in listing.values() if isinstance(v, list)]
    reduced_list_values = reduce(lambda x, y: set(x).union(y), list_values)
    types_list_values = {type(v) for v in reduced_list_values}
    assert len(types_list_values) == 1


@mark.parametrize("listing", JOINED_LISTING)
def test_listing_keys_are_lower(listing):
    """Ensure that all keys in a joined listing are in lowercase."""
    orig_keys = set(listing.keys())
    lowered_keys = map(lambda x: x.lower(), orig_keys)
    assert len(orig_keys.union(lowered_keys)) == len(orig_keys)



"""
chrono24
~~~~~~~~
"""

import sys

from chrono24.api import chrono24


class Chrono24ModuleCall:
    """A class designed to provide a callable interface for accessing the chrono24 module."""

    def __call__(self, query):
        """Call method to execute the Chrono24ModuleCall class instance as a function.

        Args:
            query (str): A string representing the query to be passed to chrono24.

        Returns:
            chrono24: An object or result returned by the chrono24 instance.
        """
        return chrono24(query)


sys.modules[__name__] = Chrono24ModuleCall()

__version__ = "0.1.2"
